#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#ifdef WIN32
 #include "windows.h"
#endif
#include <GL/gl.h>

#include "clsApplication.h"

clsApplication::clsApplication()
{
  mN=1;
  mD=1;
  mDelta=90.0f;

  glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
  glLineWidth(1.0f);
  glEnable(GL_LINE_SMOOTH);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  glPointSize(4.0f);
}

void clsApplication::openAndParseFile(char* pFile)
{
  FILE* lFile=fopen(pFile,"r");
  if (!lFile) {
    printf("Couldn't open specified file!\n");
    return;
  }

  char lBuffer[512];

  while (fscanf(lFile,"%511s ",lBuffer)!=EOF)
  {
    if (strchr(lBuffer,'=')) { //variable
      char lVar[32];
      memcpy(lVar,lBuffer,strchr(lBuffer,'=')-lBuffer);
      lVar[strchr(lBuffer,'=')-lBuffer]=0;
      if (strcmp(lVar,"n")==0)
        sscanf(strchr(lBuffer,'=')+1,"%u",&mN);
      else if (strcmp(lVar,"d")==0)
        sscanf(strchr(lBuffer,'=')+1,"%u",&mD);
      else if (strcmp(lVar,"delta")==0)
        sscanf(strchr(lBuffer,'=')+1,"%f",&mDelta);
      else
        printf("Unable to parse '%s' (unknown variable)\n",lBuffer);
    } else if (strstr(lBuffer,"->")) { //production rule
      string lConL, lConR;
      string lPre=string(lBuffer,strstr(lBuffer,"->")-lBuffer);
      string lSuc=string(strstr(lBuffer,"->")+2,strlen(lBuffer)-(strstr(lBuffer,"->")-lBuffer+2));

      if (lPre.find('<')!=string::npos) {
        lConL=lPre.substr(0,lPre.find('<'));
        lPre.erase(0,lPre.find('<')+1);
      }
      if (lPre.find('>')!=string::npos) {
        lConR=lPre.substr(lPre.find('>')+1);
        lPre.erase(lPre.find('>'));
      }
      if (lPre.length()<1) {
        printf("Error parsing '%s'\n",lBuffer);
        continue;
      }

      Production lProd;
      lProd.contextl=lConL;
      lProd.contextr=lConR;
      lProd.predecessor=lPre;
      lProd.rule=lConL+lPre+lConR;
      lProd.successor=lSuc;

      mProductions.push_back(lProd);
    } else { //something else -> hopefully axiom :D
      mAxiom=string(lBuffer);
    }
  }

  printf("Finished parsing file: \n");
  printf("  n=%u, d=%u, delta=%f\n  %s\n",mN,mD,mDelta,mAxiom.c_str());
  for (vector<Production>::iterator it=mProductions.begin();it!=mProductions.end();it++)
  {
    printf("  ");
    if (it->contextl.length()) printf("%s<",it->contextl.c_str());
    printf("%s",it->predecessor.c_str());
    if (it->contextr.length()) printf(">%s",it->contextr.c_str());
    printf("->%s\n",it->successor.c_str());
  }
  printf("-----------------------\n");

  fclose(lFile);
}

unsigned int clsApplication::changeN(int pChange)
{
  if (pChange<0 && mN==0) return mN;
  return mN+=pChange;
}
void clsApplication::produce()
{
  unsigned int lReplaces=0;
  string lProduct;
  mProduct=mAxiom;
  for(unsigned int i=0;i<mN;i++)
  {
    printf("%u ",i+1);
    while(mProduct.length())
    {
      vector<Production>::iterator it;
      for (it=mProductions.begin();it!=mProductions.end();++it) //finds suitable production rule
      {
        if (mProduct.compare(0,it->rule.length(),it->rule)!=0) continue; //this one is not useful right now

        lProduct.append(it->contextl); //append translation to lProduct
        lProduct.append(it->successor);
        lProduct.append(it->contextr);

        mProduct.erase(0,it->rule.length()); //remove letters from mProduct
        ++lReplaces;
        break;
      }
      if (it==mProductions.end()) { //no rule can be used -> just copy the letter
        lProduct.append(1,mProduct[0]);
        mProduct.erase(0,1);
      }
    }
    mProduct=lProduct; //move translation from temporary variable to mProduct
    lProduct.clear();
    //printf("%u.  %s\n",i+1,mProduct.c_str());
  }
  printf("\n");
  printf("Production done: %u replaces done, final product has %u letters\n",lReplaces,mProduct.length());
  printf("----------------\n");
}

void clsApplication::render(bool pDraw, bool pInfo)
{
  unsigned int lLines=0;
  mTurtle.x=0;
  mTurtle.y=0;
  mTurtle.alpha=270;

  float lBoundaries[4];
  lBoundaries[0]=lBoundaries[2]=FLT_MAX;
  lBoundaries[1]=lBoundaries[3]=FLT_MIN;  
  if (pDraw) {
    glClear(GL_COLOR_BUFFER_BIT);
    glLoadIdentity();
    glColor3f(1,0,0);
    glBegin(GL_POINTS);
      glVertex2f(0,0);
    glEnd();
    glColor3f(0,0,0);
  }
  for (unsigned int i=0;i<mProduct.length();++i)
  {
    TurtleState lOldState=mTurtle;
    switch (mProduct[i])
    {
      case 'f':
        mTurtle.x+=mD*cos(mTurtle.alpha*PI/180.0f);
        mTurtle.y+=mD*sin(mTurtle.alpha*PI/180.0f);
        break;
      case 'F':
        mTurtle.x+=mD*cos(mTurtle.alpha*PI/180.0f);
        mTurtle.y+=mD*sin(mTurtle.alpha*PI/180.0f);
        if (pDraw) {
          glBegin(GL_LINES);
            glVertex2f(lOldState.x,lOldState.y);
            glVertex2f(mTurtle.x,mTurtle.y);
          glEnd();
          ++lLines;
        } else {
          if (mTurtle.x<lBoundaries[0]) lBoundaries[0]=mTurtle.x;
          if (mTurtle.x>lBoundaries[1]) lBoundaries[1]=mTurtle.x;
          if (mTurtle.y<lBoundaries[2]) lBoundaries[2]=mTurtle.y;
          if (mTurtle.y>lBoundaries[3]) lBoundaries[3]=mTurtle.y;
        }
        break;
      case '+':
        mTurtle.alpha+=mDelta;
        mTurtle.alpha-=mTurtle.alpha > 360.0f ? 360.0f : 0.0f;
        break;
      case '-':
        mTurtle.alpha-=mDelta;
        mTurtle.alpha+=mTurtle.alpha < 360.0f ? 360.0f : 0.0f;
        break;
      case '|':
        mTurtle.alpha+=180;
        mTurtle.alpha-=mTurtle.alpha > 360.0f ? 360.0f : 0.0f;
        break;
      case '[':
        mTurtleStack.push(mTurtle);
        break;
      case ']':
        if (mTurtleStack.size()==0)
        {
          printf("Stack error !"); return;
        }
        mTurtle=mTurtleStack.top();
        mTurtleStack.pop();
        break;
      default:
        break;
    };
  }
  if (pInfo) {
    printf("Rendering done: %u lines were drawn\n",lLines);
    printf("--------------------------------\n");
  }
  if (!pDraw) {
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    float lXsize=(abs(lBoundaries[0])+abs(lBoundaries[1]));
    float lYsize=(abs(lBoundaries[2])+abs(lBoundaries[3]));
    if (lXsize>lYsize) {
      lBoundaries[2]-=(lXsize-lYsize)/2;
      lBoundaries[3]+=(lXsize-lYsize)/2;
    } else {
      lBoundaries[0]-=(lYsize-lXsize)/2;
      lBoundaries[1]+=(lYsize-lXsize)/2;
    }
    float lMargin=(abs(lBoundaries[0])+abs(lBoundaries[1]))*0.05f;
    glOrtho(lBoundaries[0]-lMargin,lBoundaries[1]+lMargin,lBoundaries[3]+lMargin,lBoundaries[2]-lMargin,-1.0f,1.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
  }
}
